﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.BinaryVectorEncoding.SolutionModel {
  public enum ModelBiasOptions {
    Unbiased,
    RankBiased,
    FitnessBiased
  };

  [Item("Univariate Model Training Operator", "Adds a univariate solution sampling model to the scope.")]
  [StorableClass]
  public class UnivariateModelTrainingOperator : InstrumentedOperator, IStochasticOperator, ISingleObjectiveOperator {

    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }

    public ILookupParameter<BoolValue> MaximizationParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["Maximization"]; }
    }

    public IScopeTreeLookupParameter<BinaryVector> BinaryVectorParameter {
      get { return (IScopeTreeLookupParameter<BinaryVector>)Parameters["BinaryVector"]; }
    }

    public IScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }

    public ILookupParameter<UnivariateModel> ModelParameter {
      get { return (ILookupParameter<UnivariateModel>)Parameters["Model"]; }
    }

    public IValueParameter<EnumValue<ModelBiasOptions>> ModelBiasParameter {
      get { return (IValueParameter<EnumValue<ModelBiasOptions>>)Parameters["ModelBias"]; }
    }

    public ModelBiasOptions ModelBias {
      get { return ModelBiasParameter.Value.Value; }
      set { ModelBiasParameter.Value.Value = value; }
    }

    [StorableConstructor]
    protected UnivariateModelTrainingOperator(bool deserializing) : base() { }
    protected UnivariateModelTrainingOperator(UnivariateModelTrainingOperator original, Cloner cloner)
      : base(original, cloner) { }
    public UnivariateModelTrainingOperator() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use for sampling from the model."));
      Parameters.Add(new LookupParameter<BoolValue>("Maximization", "Whether or not solution qualities are to be maximized."));
      Parameters.Add(new ScopeTreeLookupParameter<BinaryVector>("BinaryVector", "The population of solutions to create the model."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", "The fitness values of the solutions."));
      Parameters.Add(new LookupParameter<UnivariateModel>("Model", "The model that is being trained by this operator."));
      Parameters.Add(new ValueParameter<EnumValue<ModelBiasOptions>>("ModelBias", "The type of bias that is used to created the model.", new EnumValue<ModelBiasOptions>(ModelBiasOptions.Unbiased)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new UnivariateModelTrainingOperator(this, cloner);
    }

    public override IOperation InstrumentedApply() {
      var random = RandomParameter.ActualValue;
      var population = BinaryVectorParameter.ActualValue;
      var biasType = ModelBiasParameter.Value.Value;
      UnivariateModel model = null;
      switch (biasType) {
        case ModelBiasOptions.Unbiased:
          model = UnivariateModelTrainer.TrainUnbiased(random, population);
          break;
        case ModelBiasOptions.RankBiased: {
            var qualities = QualityParameter.ActualValue;
            var maximization = MaximizationParameter.ActualValue.Value;
            model = UnivariateModelTrainer.TrainWithRankBias(random, maximization, population,
              qualities.Select(x => x.Value));
          }
          break;
        case ModelBiasOptions.FitnessBiased: {
            var qualities = QualityParameter.ActualValue;
            var maximization = MaximizationParameter.ActualValue.Value;
            model = UnivariateModelTrainer.TrainWithFitnessBias(random, maximization, population,
              qualities.Select(x => x.Value));
          }
          break;
        default: throw new InvalidOperationException(string.Format("Unknown bias type {0}", biasType));
      }
      ModelParameter.ActualValue = model;
      return base.InstrumentedApply();
    }
  }
}
